#include <nds.h>

#include <FLAC/stream_decoder.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <ipc2.h>

#define SOUND_NEED_BUFFER

#include "sound.h"
#include "file.h"

#define FLAC_BUFFER_SIZE (2 * 65536) /* maximum blocksize according to flac specs */

FLAC__StreamDecoder *d;
static char *readbufL, *readbufR;
static int lastbuf, lastpos;
static u32 position;

FLAC__StreamDecoderWriteStatus flac_write(const FLAC__StreamDecoder *d, const FLAC__Frame *frame, const FLAC__int32 *const buffer[], void *client_data) {
	int j;
	s16 *dstL, *dstR;

	if(sound_channels == 0) {
		sound_channels = frame->header.channels;
		sound_frequency = frame->header.sample_rate;
		sound_bps = frame->header.bits_per_sample / 8;
	}

	dstL = ((s16 *) readbufL);
	dstR = ((s16 *) readbufR);
	for(j = 0; j < frame->header.blocksize; j++) {
		dstL[j] = (s16) buffer[0][j];
		if(sound_channels == 2)
			dstR[j] = (s16) buffer[1][j];
	}

	lastbuf = frame->header.blocksize;

	return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

void flac_metadata(const FLAC__StreamDecoder *decoder, const FLAC__StreamMetadata *metadata, void *client_data) {
}

void flac_error(const FLAC__StreamDecoder *decoder, FLAC__StreamDecoderErrorStatus status, void *client_data) {
	state = FINISHING;
}

void flac_update(void) {
	int i;
	int oldval;

	s8 *src8, *dst8;
	s16 *src16, *dst16;

	while(buffer_samples < buffer_size) {
		if(lastbuf == 0) {
			FLAC__StreamDecoderState s;

			while(lastbuf == 0) {
				s = FLAC__stream_decoder_get_state(d);
				if(s == FLAC__STREAM_DECODER_END_OF_STREAM) {
					state = FINISHING;
					return;
				}

				if(FLAC__stream_decoder_process_single(d) != 1) {
					state = ERROR;
					return;
				}
			}
			lastpos = 0;
		}

		/* we can't be interrupted here or the sound will be corrupt
		 * this shold be very fast, we're only copying the buffers */
		oldval = REG_IME;
		REG_IME = 0;
		if(sound_bps == 1) {
			src8 = ((s8 *) readbufL) + lastpos;
			dst8 = ((s8 *) pcmbufL) + buffer_end;
			for(i = 0; i < lastbuf && i < buffer_size - buffer_end && buffer_samples + i < buffer_size; i++)
				dst8[i] = src8[i];

			if(sound_channels == 2) {
				src8 = ((s8 *) readbufR) + lastpos;
				dst8 = ((s8 *) &pcmbufR) + buffer_end;

				for(i = 0; i < lastbuf && i < buffer_size - buffer_end && buffer_samples + i < buffer_size; i++)
					dst8[i] = src8[i];
			}
		} else if(sound_bps == 2) {
			src16 = ((s16 *) readbufL) + lastpos;
			dst16 = ((s16 *) pcmbufL) + buffer_end;
			for(i = 0; i < lastbuf && i < buffer_size - buffer_end && buffer_samples + i < buffer_size; i++)
				dst16[i] = src16[i];

			if(sound_channels == 2) {
				src16 = ((s16 *) readbufR) + lastpos;
				dst16 = ((s16 *) pcmbufR) + buffer_end;

				for(i = 0; i < lastbuf && i < buffer_size - buffer_end && buffer_samples + i < buffer_size; i++)
					dst16[i] = src16[i];
			}
		}

		lastbuf -= i;
		lastpos += i;
		buffer_end += i;
		buffer_end %= buffer_size;
		buffer_samples += i;
		position += i;

		REG_IME = oldval;
	}
}

void flac_stop(void) {
	FLAC__stream_decoder_finish(d);

	free(readbufL);
	free(readbufR);
}

int flac_play(struct media *m) {
	format = UNKNOWN;

	readbufL = (char *) malloc(FLAC_BUFFER_SIZE * 2);
	readbufR = (char *) malloc(FLAC_BUFFER_SIZE * 2);
	lastpos = lastbuf = 0;
	position = 0;

	d = FLAC__stream_decoder_new();
	FLAC__stream_decoder_init_file(d, m->path, flac_write, flac_metadata, flac_error, NULL);

	FLAC__stream_decoder_process_until_end_of_metadata(d);

	sound_channels = 0;

	flac_update();

	if(state == FINISHING) {
		flac_stop();
		state = ERROR;
		return 1;
	}

	format = FLAC;

	return 0;
}

void flac_flush(void) {
	lastbuf = lastpos = 0;
}

u32 flac_size(void) {
	return FLAC__stream_decoder_get_total_samples(d);
}

u32 flac_position(void) {
	return position;
}

void flac_seek(u32 pos) {
	FLAC__stream_decoder_seek_absolute(d, pos);
	position = pos;
}
